package eu.connect.photo.client.backend.lime;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import lights.adapters.Tuple;
import lights.interfaces.ITuple;
import lime.AgentLocation;
import lime.HostLocation;
import lime.IllegalTupleSpaceNameException;
import lime.LimeServer;
import lime.LimeServerID;
import lime.LimeSystemTupleSpace;
import lime.LimeTemplate;
import lime.LimeTuple;
import lime.LimeTupleSpace;
import lime.StationaryAgent;
import lime.TupleSpaceEngineException;
import eu.connect.photo.core.Photo;
import eu.connect.photo.core.PhotoMetaData;
import eu.connect.photo.core.PhotoQuery;
import eu.connect.photo.core.PhotoSharing;

public class LIMEBackEnd extends StationaryAgent implements PhotoSharing {

	private static final long serialVersionUID = -427603639625557828L;
	private static final String METADATA_TS_NAME = "CONNECTMetaDataTS";
	private static final String DATA_TS_NAME = "CONNECTDataTS";

	private LimeSystemTupleSpace system;
	private LimeTupleSpace metadataTupleSpace;
	private LimeTupleSpace dataTupleSpace;
	private boolean stopped;
	private List<LimeAction> actionsToDo;

	public LIMEBackEnd() {
		super(LIMEBackEnd.class.getName());
		this.actionsToDo = new LinkedList<LIMEBackEnd.LimeAction>();
		this.stopped = false;
	}

	public LIMEBackEnd(String name) {
		super(name);
		this.actionsToDo = new LinkedList<LIMEBackEnd.LimeAction>();
		this.stopped = false;
	}

	public LimeTupleSpace getMetaData() {
		return metadataTupleSpace;
	}

	public LimeTupleSpace getData() {
		return dataTupleSpace;
	}

	/**
	 * Implement the retrieval of a photo from the tuple space.
	 * 
	 * @param photoId
	 *            the id of the photo to retrieve
	 * @return the related Photo object
	 */
	private Photo doGet(int photoId) {
		Photo result = null;
		
		print("Retrieving Photo (id=" + photoId + ")\n");

		// Extract the meta-data from the photo id
		List<PhotoMetaData> allMetaData = doSearchMetaData(new PhotoQuery(photoId));
		if (allMetaData.isEmpty()) {
			print("Error! Unable to locate the picture. The provider may have departed.\n");
			
		} else if (allMetaData.size() > 1) {
			print("Error! Duplicated Photo ID!\n");
		
		} else  {
			// Extract the content of the photo
			byte[] data = doSearchData(photoId);
			if (data != null) {
				result = new Photo(allMetaData.get(0), data);
				print("Photo retrieved!\n");
			}
		}
		
		return result;
	}

	/**
	 * Implement the Sharing process. First we retrieve all the meta-data from
	 * the tuple in order to calculate the new ID for the photo to share. Then,
	 * we publish the meta-data and data.
	 * 
	 * @param photo
	 *            the photo to share.
	 */
	private void doShare(Photo photo) {
		print("Sharing a new picture ... \n");

		// Calculate the id of the picture. Browse the existing pictures and
		List<PhotoMetaData> allPhotos = doSearchMetaData(new PhotoQuery());
		int id = calculateNextID(allPhotos);
		photo.getMetaData().setId(id);
		print(" - Generation of a new ID: #" + id + "\n");

		try {
			// Share the meta-data
			print(" - Publishing meta-data \n");
			ITuple metaDataTuple = metadataToTuple(photo.getMetaData());
			metadataTupleSpace.out(metaDataTuple);

			// Share the data
			print(" - Publishing content \n");
			ITuple dataTuple = dataToTuple(photo.getMetaData().getId(), photo
					.getRawData());
			dataTupleSpace.out(dataTuple);

			print("Sharing complete.\n");

		} catch (TupleSpaceEngineException e) {
			print("Error while writing in the tuple spaces!\n");
			e.printStackTrace();

		}

	}

	/**
	 * Query the federated tuple space in order to retrieve all the meta-data of
	 * matching a given query. This should basically be done using the "rdg"
	 * command, but since the current LIME implementation does not support
	 * unconstrained bulk operations, we first retrieve all the hosts from the
	 * LIME system tuple space and then we query each of them consequently.
	 * 
	 * @param query
	 *            the query to resolve
	 * @return the list of matching meta-data
	 */
	private List<PhotoMetaData> doSearchMetaData(PhotoQuery query) {
		List<PhotoMetaData> result = new LinkedList<PhotoMetaData>();

		print("Searching for meta-data matching the query " + query + "\n");

		ITuple template = queryToTemplate(query);

		List<LimeServerID> allHosts = getAllHosts();
		for (LimeServerID host : allHosts) {
			try {
				LimeTuple[] allTuples = metadataTupleSpace
						.rdg(new LimeTemplate(new HostLocation(host),
								AgentLocation.UNSPECIFIED, null, template));
				if (allTuples != null) {
					print(" - Searching in " + host + " ... " + allTuples.length + " tuple found!\n");
					for (LimeTuple tuple : allTuples) {
						result.add(tupleToMetaData(tuple));
					}
				} else {
					print(" - Searching in " + host + " ... No tuple found.\n");
				}

			} catch (TupleSpaceEngineException e) {
				print("Error while retrieving information from " + host + "\n");
			}
		}

		return result;
	}
	
	
	private byte[] doSearchData(int photoId) {
		print("Searching for content of the picture #" + photoId + "\n");

		// Create a template to query the hosts
		ITuple template = new Tuple();
		template.addActual(photoId);
		template.addFormal(new byte[2].getClass());
		
		// Searching through all the available host
		byte[] result = null;
		List<LimeServerID> allHosts = getAllHosts();
		Iterator<LimeServerID> iterator = allHosts.iterator();
		while(iterator.hasNext() && result == null) {
			LimeServerID host = iterator.next();
			try {
				LimeTuple[] allTuples = dataTupleSpace
						.rdg(new LimeTemplate(new HostLocation(host),
								AgentLocation.UNSPECIFIED, null, template));
				if (allTuples != null) {
					if (allTuples.length > 1) {
						print(" - Searching in " + host + " ... Error, Duplicated ID!\n");
					} else {
						print(" - Searching in " + host + " ... Found!\n");
						result = (byte[]) allTuples[0].getTuple().get(1).getValue();
					}
				} else {
					print(" - Searching in " + host + " ... No tuple found.\n");
				}

			} catch (TupleSpaceEngineException e) {
				print("Error while retrieving information from " + host + "\n");
			}
		}

		return result;
	}

	/**
	 * Calculate the id of the picture to add in the tuple. From the set of
	 * existing tuple, we search for the max id, and return its increment by one
	 * 
	 * @param existingTuples
	 *            the list of available metaData
	 * @return the id of the new picture
	 */
	private int calculateNextID(List<PhotoMetaData> existingMetaData) {
		int result = 0;
		for (PhotoMetaData pmd : existingMetaData) {
			print("ID: " + pmd.getId() );
			if (pmd.getId() > result) {
				result = pmd.getId();
			}
		}
		return result + 1;
	}

	/**
	 * Retrieve the addresses of all the host currently available in the
	 * community. This is done using the LimeSystem Tuple space.
	 * 
	 * @return
	 */
	private List<LimeServerID> getAllHosts() {
		List<LimeServerID> result = new LinkedList<LimeServerID>();

		// Create a template to query the LIME system tuple space
		ITuple hostTemplate = new Tuple();
		hostTemplate.addActual("_host");
		hostTemplate.addFormal(LimeServerID.class);

		try {
			ITuple[] allHosts = system.rdg(hostTemplate);
			if (allHosts != null) {
				for (ITuple tuple : allHosts) {
					result.add((LimeServerID) tuple.get(1).getValue());
				}
			}

		} catch (TupleSpaceEngineException e) {
			print("Error, unable to retrieve the list of host from the system tuple space!\n");
		}

		return result;
	}

	public void run() {

		try {
			// Engage the host in the LIME community
			LimeServer.getServer().engage();
			if (LimeServer.getServer().isEngaged())
				print("Entering the community!\n");

			// Create and share all the tuple spaces
			system = new LimeSystemTupleSpace();
			metadataTupleSpace = new LimeTupleSpace(METADATA_TS_NAME);
			metadataTupleSpace.setShared(true);
			dataTupleSpace = new LimeTupleSpace(DATA_TS_NAME);
			dataTupleSpace.setShared(true);
			if (metadataTupleSpace.isShared() && dataTupleSpace.isShared())
				print("Tuples spaces initialized and shared!\n");

			print("Ready! Waiting for requests ...\n");

			while (!stopped) {

				// If the queue if empty, then we wait until an action has been
				// put in the queue.
				if (actionsToDo.isEmpty()) {
					synchronized (actionsToDo) {
						try {
							actionsToDo.wait();

						} catch (InterruptedException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}

				} else {

					// We extract the action and then we execute it
					LimeAction action = actionsToDo.remove(0);
					action.execute(this);

					// Notify the others that the action is now complete
					synchronized (action) {
						action.notifyAll();
					}
				}
			}

		} catch (IllegalTupleSpaceNameException e) {
			print("Error while opening the tuple space!\n");
			e.printStackTrace();

		} catch (TupleSpaceEngineException e) {
			print("Error while opening the tuple space!\n");
			e.printStackTrace();
		}

	}

	public void shutdown() {
		// Set the value to stop
		this.stopped = true;

		// Wake the server if he was stopped
		synchronized (this.actionsToDo) {
			actionsToDo.notifyAll();
		}
	}

	/**
	 * Share a picture on the tuple space. Calling this function is blocking
	 * until the share action is performed by the LimeAgent
	 * 
	 * @param photo
	 *            the photo to share
	 */
	public void share(Photo photo) {
		// Build the related action
		ShareAction shareAction = new ShareAction(photo);
		this.actionsToDo.add(shareAction);

		// Notify others that an new action is the queue
		synchronized (actionsToDo) {
			actionsToDo.notifyAll();
		}

		// This should be a blocking call
		// Wait until the action to be complete
		synchronized (shareAction) {
			try {
				shareAction.wait();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Search for photos matching the given query, i.e. matching the tags
	 * defined in the query. Calling this function is blocking until the query
	 * as been resolved.
	 * 
	 * @param query
	 *            the query to resolve
	 * @return a list of meta data describing the picture matching the query
	 */
	public List<PhotoMetaData> search(PhotoQuery query) {
		// Create the related action
		SearchAction searchAction = new SearchAction(query);
		this.actionsToDo.add(searchAction);

		// Notify others that an new action is the queue
		synchronized (actionsToDo) {
			actionsToDo.notifyAll();
		}

		// This should be a blocking call
		// Wait until the action to be complete
		synchronized (searchAction) {
			try {
				searchAction.wait();

			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		return searchAction.getResult();
	}

	/**
	 * Retrieve a photo from its meta-data. Only the id of the photo is used to
	 * query the photo, since we suppose id unique for all pictures.
	 * 
	 * @param id
	 *            the id of the photo to retrieve
	 * @return the photo matching this particular id
	 */
	public Photo get(int id) {
		// Create the related action
		GetAction getAction = new GetAction(id);
		this.actionsToDo.add(getAction);

		// Notify others that an new action is the queue
		synchronized (actionsToDo) {
			actionsToDo.notifyAll();
		}

		// This should be a blocking call
		// Wait until the action to be complete
		synchronized (getAction) {
			try {
				getAction.wait();

			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}

		return getAction.getResult();
	}

	/**
	 * Generate a tuple to query the metadata tuplespace.
	 * 
	 * @param query
	 *            the query to translate
	 * @return a tuple that can be used as a template to search through the
	 *         meta-data tuplespace
	 */
	private ITuple queryToTemplate(PhotoQuery query) {
		ITuple result = new Tuple();

		// Handle the id
		if (query.isIdUnspecified()) {
			result.addFormal(Integer.class);			
		} else {
			result.addActual(query.getId());
		}

		// Handle the title
		if (query.isTitleUnspecified()) {
			result.addFormal(String.class);
		} else {
			result.addActual(query.getTitle());
		}

		// Handle the date
		if (query.isDateUnspecified()) {
			result.addFormal(String.class);
		} else {
			result.addActual(query.getDate());
		}

		// Handle the Location
		if (query.isLocationUnspecified()) {
			result.addFormal(String.class);
		} else {
			result.addActual(query.getLocation());
		}

		return result;
	}

	/**
	 * Convert a PhotoMetaData object in a tuple that can be stored in the
	 * tuplespace for metadata
	 * 
	 * @param metaData
	 *            the metaData object that must be converted
	 * @return a tuple equivalent to the MetaData object
	 */
	private static ITuple metadataToTuple(PhotoMetaData metaData) {
		ITuple result = new Tuple();

		result.addActual(metaData.getId());
		result.addActual(metaData.getTitle());
		result.addActual(metaData.getDate());
		result.addActual(metaData.getLocation());

		return result;
	}

	/**
	 * Convert a tuple extracted from the tuple space containing metadata to a
	 * PhotoMetaData abject. We assume that the tuple matches the format
	 * <Integer, String, String, String>
	 * 
	 * @param limeTuple
	 *            the tuple object to convert
	 * @return the resulting PhotoMetaData object
	 */
	private static PhotoMetaData tupleToMetaData(LimeTuple limeTuple) {

		Integer id = (Integer) limeTuple.getTuple().get(0).getValue();
		String title = (String) limeTuple.getTuple().get(1).getValue();
		String date = (String) limeTuple.getTuple().get(2).getValue();
		String location = (String) limeTuple.getTuple().get(3).getValue();

		return new PhotoMetaData(id, title, date, location);
	}


	/**
	 * Create a data tuple from the id of photo and its raw data
	 * 
	 * @param id
	 *            the id of the photo
	 * @param rawData
	 *            the raw data
	 * @return the related tuple to insert into the data tuple space
	 */
	private static ITuple dataToTuple(int id, byte[] rawData) {
		ITuple result = new Tuple();

		result.addActual(id);
		result.addActual(rawData);

		return result;
	}

	private static void print(String message) {
		System.out.print("LIME BackEnd: " + message);
	}

	/**
	 * Abstract a Lime action, e.g. Share a photo, Search for photo, or get a
	 * photo
	 */
	private abstract static class LimeAction {
		public abstract void execute(LIMEBackEnd limeBackEnd);

	}

	/**
	 * Encapsulates all the information of a share action
	 */
	private static class ShareAction extends LimeAction {
		private Photo photo;

		public ShareAction(Photo photo) {
			this.photo = photo;
		}

		@Override
		public void execute(LIMEBackEnd agent) {
			agent.doShare(photo);
		}

	}

	/**
	 * Encapsulate a query to the photo center
	 */
	private static class SearchAction extends LimeAction {

		private PhotoQuery query;
		private List<PhotoMetaData> result;

		public SearchAction(PhotoQuery query) {
			this.query = query;
			this.result = new LinkedList<PhotoMetaData>();
		}

		public List<PhotoMetaData> getResult() {
			return result;
		}

		@Override
		public void execute(LIMEBackEnd agent) {
			result.addAll(agent.doSearchMetaData(query));
		}

	}

	/**
	 * Encapsulate a query on a single photo
	 */
	private static class GetAction extends LimeAction {
		private int photoId;
		private Photo result;

		public GetAction(int id) {
			this.photoId = id;
		}

		public Photo getResult() {
			return this.result;
		}

		public void execute(LIMEBackEnd agent) {
			result = agent.doGet(photoId);
		}
	}

}
