package eu.connect.photo.client.console;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

import lime.ILimeAgent;
import lime.LimeException;
import lime.LimeServer;
import eu.connect.photo.client.backend.lime.LIMEBackEnd;
import eu.connect.photo.client.backend.rmi.RmiBackEnd;
import eu.connect.photo.client.backend.soap.SOAPBackEnd;
import eu.connect.photo.client.console.command.Command;
import eu.connect.photo.client.console.command.CommandBuilder;
import eu.connect.photo.client.console.command.IllegalCommand;
import eu.connect.photo.core.Photo;
import eu.connect.photo.core.PhotoMetaData;
import eu.connect.photo.core.PhotoQuery;
import eu.connect.photo.core.PhotoSharing;

public class Console {

	public static final String LIME_CLIENT = "lime";
	public static final String SOAP_CLIENT = "soap";
	public static final String RMI_CLIENT = "rmi";

	public static final String PROMPT = "PhotoSharing> ";

	private CommandBuilder builder;
	private PhotoSharing backEnd;
	private boolean stopped;
	private List<Command> history;

	/**
	 * Create a new PhotoSharingClient. No back end are defined by default.
	 * 
	 * @see setBackEnd
	 */
	public Console() {
		builder = CommandBuilder.getInstance();
		this.history = new LinkedList<Command>();
		this.stopped = false;
	}

	/**
	 * Change the back end of the client
	 * 
	 * @param backEnd
	 *            the new back end to use
	 */
	public void setBackEnd(PhotoSharing backEnd) {
		this.backEnd = backEnd;
	}

	/**
	 * Return the back currently used by the client
	 * 
	 * @return the backend object used by the client
	 */
	public PhotoSharing getBackEnd() {
		return this.backEnd;
	}

	/**
	 * Build a Photo object from the name of a file, the list of tags describing
	 * the photo. Delegate to the back-end the real sharing of the photo object.
	 * 
	 * @param fileToShare
	 *            the path to the file to share
	 * @param title
	 *            the title of photo
	 * @param date
	 *            the date where the photo was shot
	 * @param location
	 *            the place where the photo was shot
	 */
	public void share(String fileToShare, String title, String date,
			String location) {
		File file = new File(fileToShare);

		try {
			Photo photo = new Photo(new PhotoMetaData(title, date, location),
					file);
			backEnd.share(photo);

		} catch (FileNotFoundException e) {
			System.out
					.println("Error: The file you specified cannot be found: ("
							+ fileToShare + ")");

		} catch (IOException e) {
			System.out
					.println("Unexpected IO error while reading the file from the disk. Please retry later.");

		}

	}

	/**
	 * Retrieve a given photo from its ID
	 * 
	 * @param id
	 *            the id of the photo to retrieve
	 * @param fileToCreate
	 *            the name/path of the the file to create
	 */
	public void savePhoto(int id, String fileToCreate) {
		Photo photo = backEnd.get(id);
		if (photo == null) {
			System.out.println("Picture ID not found!");
			System.out.println("There is no picture matching the id \"" + id
					+ "\".");
			System.out
					.println("Please use the SEARCH command to get the list of pictures and their id");

		} else {
			try {
				photo.storeAs(fileToCreate);

			} catch (IOException e) {
				System.out
						.println("Unexpected error while writing the picture on the disk. Please retry later");
			}
		}
	}

	/**
	 * Search for picture matching the specific meta-data (title, date,
	 * location)
	 * 
	 * @param title
	 *            the title of the picture
	 * @param date
	 *            the date where the picture was shot
	 * @param location
	 *            the place where the picture was shot
	 * @param tags
	 *            a list of additional tags describing the picture
	 */
	public void search(String title, String date, String location,
			List<String> tags) {
		PhotoQuery query = new PhotoQuery(title, date, location);
		List<PhotoMetaData> result = backEnd.search(query);
		if (!result.isEmpty()) {
			System.out.println("Photo(s) matching " + query + ":");
			System.out.println(String.format("%3s %15s %15s %15s", "ID",
					"Title", "Date", "Location"));
			for (PhotoMetaData metaData : result) {
				System.out.println(metaData);
			}

		} else {
			System.out.println("No photo match the query " + query);
		}
	}

	public void exit() {
		System.out.println("That's all folks!");
		this.stopped = true;
	}

	public void history() {
		System.out.println("Last commands:");
		int counter = history.size();
		for (Command command : this.history) {
			System.out.println(" - " + counter + ": " + command);
			counter--;
		}
	}

	public void help() {
		System.out.println("The available commands are:");
		System.out.println(" - \"exit\": Quit the PhotoSharingClient");
		System.out
				.println(" - \"save X as file.jpg\": Retrieve the photo whose ID's equals to X and save it as file.jpg");
		System.out
				.println(" - \"help\": Display the list of available commands");
		System.out
				.println(" - \"history\": Show the commands previously entered.");
		System.out
				.println(" - \"search title date location\": Search for picture matching the provided tag list");
		System.out
				.println(" - \"share myFile.jpg title location date\": Make a photo available to other users");
	}

	public void start() {
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				System.in));
		this.stopped = false;
		while (!stopped) {
			System.out.print("\n" + PROMPT);
			try {
				String commandText = readCommandFromConsole(reader);
				Command command = builder.build(commandText);
				history.add(command);
				command.executeOn(this);

			} catch (IllegalCommand e) {
				System.out.println(e);

			}
		}
	}

	/**
	 * Read a text line from the console.
	 * 
	 * @param reader
	 *            the stream/reader where the text line must be read
	 * @return the line of text which has been read
	 * @throws IllegalCommand
	 *             if an unexpected IO error occurs while reading the line of
	 *             text
	 */
	private String readCommandFromConsole(BufferedReader reader)
			throws IllegalCommand {

		String commandText = null;
		try {
			commandText = reader.readLine();
			commandText.trim();

		} catch (IOException ioe) {
			IllegalCommand error = new IllegalCommand("\"" + commandText
					+ "\"(maybe not available)");
			error.setError("Unexpected IO error while reading the command");
			error.setAdvice("Try reentering the command.");
			throw error;
		}

		return commandText;
	}

	/**
	 * Entry Point of the Client application. Provide a textual interface to the
	 * photo sharing application. Could be connected to various "back-ends"
	 * implementing a given technology, such as SOAP, RMI, LIME, etc
	 * 
	 * @param args
	 *            a string specifying the back end to use, either "LIME" or
	 *            "SOAP"
	 */
	public static void main(String args[]) {
		System.out.println("PhotoSharing Client v0.1");
		System.out.println("(c) CONNECT Project - 2010");
		System.out.println("----");

		// Create the Client
		Console client = new Console();

		// Analyze the command parameters and configure the client accordingly
		if (args.length > 1) {
			System.out.println("Usage: $>java PhotoSharingClient [LIME|SOAP]");
			System.exit(1);
		}

		// Extract the parameters
		String backEndName = LIME_CLIENT;
		if (args.length == 1) {
			backEndName = args[0];
		}

		if (backEndName.equals(SOAP_CLIENT)) {
			client.setBackEnd(new SOAPBackEnd());

		} else if (backEndName.equals(RMI_CLIENT)) {
			client.setBackEnd(new RmiBackEnd());

		} else if (backEndName.equals(LIME_CLIENT)) {
			startLIME(client);

		} else {
			System.out
					.println("Error: Unsupported backend. Only SOAP or LIME are currently supported.");
			System.out.println("Usage: $>java PhotoSharingClient [LIME|SOAP]");
			System.exit(1);
		}

		// Start the client
		client.start();

		// The client was exited
		if (backEndName.equals(SOAP_CLIENT)) {
			// TODO

		} else if (backEndName.equals(LIME_CLIENT)) {
			stopLIME(client);

		}

	}

	/**
	 * Configure the lime Back end and connect the client with it
	 * 
	 * @param client
	 *            the client whose backend must be configure with a LIME backend
	 */
	private static void startLIME(Console client) {
		try {
			// Start the LIME server
			new lime.util.Launcher()
					.launch(new String[] { }, 0);
			// Load the agent
			ILimeAgent agent = LimeServer.getServer().loadAgent(
					LIMEBackEnd.class.getName(), new String[] {});
			client.setBackEnd((LIMEBackEnd) agent);

		} catch (LimeException le) {
			System.out.println("Trouble Loading the agent");
			le.printStackTrace();
		}
	}

	private static void stopLIME(Console client) {
		LIMEBackEnd backEnd = (LIMEBackEnd) client.getBackEnd();
		backEnd.shutdown();
	}
}
