package eu.connect.photo.client.console.command;

public class CommandBuilder {
	public static final String HELP_KW = "help";
	public static final String EXIT_KW = "exit";
	public static final String HISTORY_KW = "history";
	public static final String SHARE_KW = "share";
	public static final String SAVE_KW = "save";
	public static final String AS_KW = "as";
	public static final Object SEARCH_KW = "search";

	private static CommandBuilder instance;

	public static CommandBuilder getInstance() {
		if (instance == null)
			instance = new CommandBuilder();
		return instance;
	}

	private CommandBuilder() {

	}

	/**
	 * Build a command from a string. Parse the text and instantiate the
	 * relevant sub class of the command while extracting the related
	 * parameters.
	 * 
	 * @param commandText
	 *            the text of the command to parse
	 * @return a command object properly instantiated
	 * @throws IllegalCommand
	 *             if the text command does not match the command syntax
	 */
	public Command build(String commandText) throws IllegalCommand {
		// read the first token
		Command result = null; 

		String[] tokens = commandText.split("\\s");
		if (tokens[0].equals(EXIT_KW)) {
			result = buildExit(commandText, tokens);
 
		} else if (tokens[0].equals(HELP_KW)) {
			result = buildHelp(commandText, tokens);

		} else if (tokens[0].equals(HISTORY_KW)) {
			result = buildHistory(commandText, tokens);

		} else if (tokens[0].equals(SEARCH_KW)) {
			result = buildSearchCommand(commandText, tokens);

		} else if (tokens[0].equals(SHARE_KW)) {
			result = buildShare(commandText, tokens);

		} else if (tokens[0].equals(SAVE_KW)) {
			result = buildGet(commandText, tokens);

		} else {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("Unknown command: \"" + tokens[0] + "\"");
			error.setAdvice("Enter \"help\" to get the list of available commands and their usage.");
			throw error;
		}

		return result;
	}

	
	private Command buildSearchCommand(String commandText, String[] tokens)
			throws IllegalCommand {
		SearchCommand result = new SearchCommand();
		if (tokens.length < 4) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The SEARCH command requires at least 2 parameters.");
			error.setAdvice("Enter \"help\" to get the correct usage of the SEARCH command.");
			throw error;
		} else {
			result.setTitle(tokens[1]);
			result.setDate(tokens[2]);
			result.setLocation(tokens[3]);
			for (int index = 4; index < tokens.length; index++) {
				result.addTag(tokens[index]);
			}
		}
		return result;
	}

	private Command buildGet(String commandText, String[] tokens)
			throws IllegalCommand {
		SaveCommand result = new SaveCommand();
		if (tokens.length != 4) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The GET command requires 4 parameter.");
			error.setAdvice("Enter \"help\" to get the correct usage of the GET command.");
			throw error;
		} else {
			result.setID(Integer.parseInt(tokens[1]));
			if (tokens[2].equals(AS_KW)) {
				result.setFileToCreate(tokens[3]);
			} else {
				IllegalCommand error = new IllegalCommand(commandText);
				error.setError("The GET command requires you to specify the name of the file to create");
				error.setAdvice("Enter \"help\" to get the correct usage of the GET command.");
				throw error;		
			}
		}
		return result;
	}

	private Command buildShare(String commandText, String[] tokens)
			throws IllegalCommand {
		ShareCommand result = new ShareCommand();
		if (tokens.length < 5) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The SHARE command requires at least 5 parameters.");
			error.setAdvice("Enter \"help\" to get the correct usage of the SHARE command.");
			throw error;
		} else {
			result.setFileToShare(tokens[1]);
			result.setTitle(tokens[2]);
			result.setDate(tokens[3]);
			result.setLocation(tokens[4]);
		}
		return result;
	}

	private Command buildHistory(String commandText, String[] tokens)
			throws IllegalCommand {
		Command result = new HistoryCommand();
		if (tokens.length > 1) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The HISTORY command does not accept any parameter.");
			error.setAdvice("Enter \"help\" to get the correct usage of the HISTORY command.");
			throw error;
		}
		return result;
	}

	private Command buildHelp(String commandText, String[] tokens)
			throws IllegalCommand {
		Command result = new HelpCommand();
		if (tokens.length > 1) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The HELP command does not accept any parameter.");
			error.setAdvice("Enter \"help\" to get the correct usage of the HELP command.");
			throw error;
		}
		return result;
	}

	private Command buildExit(String commandText, String[] tokens)
			throws IllegalCommand {
		Command result = new ExitCommand();
		if (tokens.length > 1) {
			IllegalCommand error = new IllegalCommand(commandText);
			error.setError("The EXIT command does not accept any parameter.");
			error.setAdvice("Enter \"help\" to get the correct usage of the EXIT command.");
			throw error;
		}
		return result;
	}

}
