 package eu.connect.photo.client.gui;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import javax.swing.JFrame;

import lime.AgentCreationException;
import lime.ILimeAgent;
import lime.LimeServer;
import eu.connect.photo.client.backend.lime.LIMEBackEnd;
import eu.connect.photo.client.backend.rmi.RmiBackEnd;
import eu.connect.photo.core.PhotoSharing;

/**
 * This is the main class of the client GUI. As the text-based client, the GUI
 * also forwards the user command to a back-end, which deals with a given
 * protocol (SOAP, LIME, RMI, etc.)
 */
public class GUI extends JFrame {

	public static final String LIME_CLIENT = "lime";
	public static final String RMI_CLIENT = "rmi";
	public static final String SOAP_CLIENT = "soap";

	private static final long serialVersionUID = 5522884412020457030L;

	private PhotoSharing backend;

	/**
	 * Create the PhotoSharingGUI with a given backend object
	 * 
	 * @param backend
	 *            the back-end object handling the commands
	 */
	public GUI(PhotoSharing backend) {
		super("Photo Sharing");
		setMinimumSize(new Dimension(500,200));
		setResizable(false);
		
		this.backend = backend;
		
		this.getContentPane().setLayout(new GridBagLayout());
		this.add(new SharePanel(this), new GridBagConstraints(0, 0, 1, 1,
				0, 0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		this.add(new QueryPanel(this), new GridBagConstraints(0, 1, 1, 1,
				0, 0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
	}

	/**
	 * @return the back-end object used by the client
	 */
	public PhotoSharing getBackEnd() {
		return this.backend;
	}

	/**
	 * Change the back-end object used by the client
	 * 
	 * @param backend
	 *            the new back-end object to use
	 */
	public void setBackEnd(PhotoSharing backend) {
		this.backend = backend;
	}

	/**
	 * Entry point of GUI. The back end which must be used can be specified
	 * directly from the command line. By default, the back-end used is the LIME
	 * one.
	 * 
	 * @param args
	 *            the list of parameters of the GUI client. Usage is $> java
	 *            PhotoSharingGUI [lime|soap]
	 */
	public static void main(String args[]) {

		args = new String[1];
		args[0] = RMI_CLIENT;
			
		System.setProperty("java.security.policy", "resources/unsafe.policy");
		System.setProperty("java.rmi.server.codebase", "file://"+System.getProperty("user.dir")+"/bin/");
		
		
		// Analyze the command parameters and configure the client accordingly
		if (args.length > 1) {
			System.out.println("Usage: $>java PhotoSharingGUI [lime|soap]");
			System.exit(1);
		}

		// Extract the parameters
		String backEndName = LIME_CLIENT;
		if (args.length == 1) {
			backEndName = args[0];
		}

		GUI app = null;
		if (backEndName.equals(SOAP_CLIENT)) {
			// TODO Build the soap client correctly
			
		} else if (backEndName.equals(RMI_CLIENT)) {	
			app = new GUI(new RmiBackEnd());

		} else if (backEndName.equals(LIME_CLIENT)) {
			// Start the LIME server
			new lime.util.Launcher()
					.launch(new String[] {"-errors", "on", }, 0 );
			// Load the agent
			try {
				ILimeAgent agent = LimeServer.getServer().loadAgent(
						LIMEBackEnd.class.getName(), new String[] {});
				app = new GUI((LIMEBackEnd) agent);

			} catch (AgentCreationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

		} else {
			System.out
					.println("Error: Unsupported backend. Only SOAP or LIME are currently supported.");
			System.out.println("Usage: $>java PhotoSharingClient [lime|soap]");
			System.exit(1);
		}

		app.pack();
		app.setVisible(true);

	}

}
