package eu.connect.photo.server.rmi;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import eu.connect.photo.core.Photo;
import eu.connect.photo.core.PhotoMetaData;
import eu.connect.photo.core.PhotoQuery;

/**
 * Implements a Server storing photos and accessible through RMI. The server
 * receive requests but delegates them to a separate thread.
 * 
 */
public class RMIServer extends UnicastRemoteObject implements RMIPhotoSharing {

	public static final String SERVER_ID = "CONNECT_PhotoSharing";

	public static final String HOME_DIR = "data/server/";
	public static final String METADATA_FILE = "collection.xml";

	private static final long serialVersionUID = 2328561622656602858L;

	private List<PhotoMetaData> metaData;

	public RMIServer() throws RemoteException {
		super();

		try {
			System.out.println("Searching for database in: " + new File(".").getCanonicalPath().toString());
		
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		MetaDataReader handler = new MetaDataReader();
		SAXParserFactory factory = SAXParserFactory.newInstance();
		try {
			// Instantiate a parser
			XMLReader parser = factory.newSAXParser().getXMLReader();
			parser.setContentHandler(handler);
			parser.parse(HOME_DIR + METADATA_FILE);
			metaData = handler.getResult();
			System.out.println("XML Database loaded.");

		} catch (FileNotFoundException e) {
			System.out.println("No database found: Initializing a new one.");
			metaData = new LinkedList<PhotoMetaData>();

		} catch (ParserConfigurationException e) {
			System.out.println("Error while parsing the existing metadata");
			metaData = new LinkedList<PhotoMetaData>();

		} catch (SAXException e) {
			System.out.println("Ill-formed XML database: Initializing a new one.");
			metaData = new LinkedList<PhotoMetaData>();

		} catch (IOException e) {
			System.out.println("IO Error while parsing the existing metadata");
			metaData = new LinkedList<PhotoMetaData>();

		}

	}

	public Photo get(int photoId) throws RemoteException {
		System.out.println(" + Request for Photo #" + photoId);
		PhotoMetaData md = null;
		Iterator<PhotoMetaData> iterator = metaData.iterator();
		while(md == null && iterator.hasNext()) {
			PhotoMetaData current = iterator.next();
			if (current.getId() == photoId) {
				md = current;
			}
		}
		
		Photo result = null;
		try {
			result = new Photo(md, new File(HOME_DIR + "photo_" + md.getId() + ".jpg"));
		} catch (FileNotFoundException e) {
			System.out.println("Error: inconsistent database, the file related to the photo does not exist!");
			
		} catch (IOException e) {
			System.out.println("Unexpected IO error while reading the file from the disk");

		}
		
		return result;
	}

	public List<PhotoMetaData> search(PhotoQuery query) throws RemoteException {
		System.out.println("+ Searching for picture :" + query);
		List<PhotoMetaData> result = new LinkedList<PhotoMetaData>();
		for (PhotoMetaData pmd : metaData) {
			if (pmd.getTitle().equals(query.getTitle())
					|| query.getTitle().equals("*")
					|| pmd.getLocation().equals(query.getLocation())
					|| query.getLocation().equals("*")
					|| pmd.getDate().equals(query.getDate())
					|| query.getDate().equals("*")) {
				result.add(pmd);
			}
		}
		System.out.printf("%d photo(s) matching the query\n", result.size());
		return result;
	}

	private void updateDatabase() {
		try {
			PrintStream file = new PrintStream(new File(HOME_DIR + METADATA_FILE));
			file.println("<?xml version=\"1.0\" ?>");
			file.println("<photos>");
			for(PhotoMetaData md : metaData) {
				file.println("\t<photo>");
				file.printf("\t\t<id>%d</id>\n", md.getId());
				file.printf("\t\t<title>%s</title>\n", md.getTitle());
				file.printf("\t\t<date>%s</date>\n", md.getDate());
				file.printf("\t\t<location>%s</location>\n", md.getLocation());
				file.println("\t</photo>");
			}
			file.println("</photos>");
			file.close();
			System.out.println("Database Updated.");
			
		} catch (FileNotFoundException e) {
			System.out.println("Error, unable to access the database.");
		}
	}

	public void share(Photo photo) throws RemoteException {
		System.out.printf(" + Storing new picture: \n");
		int maxId = 0;
		for(PhotoMetaData pmd : metaData) {
			if (pmd.getId() > maxId) maxId = pmd.getId();
		}
		System.out.printf(" - Id: %d", maxId+1);
		System.out.printf(" - Title: %s \n", photo.getMetaData().getTitle());
		System.out.printf(" - Date: %s\n", photo.getMetaData().getDate());
		System.out.printf(" - Location: %s\n", photo.getMetaData().getLocation());
		
		photo.getMetaData().setId(maxId+1);
		metaData.add(photo.getMetaData());
		try {
			photo.storeAs(HOME_DIR + "photo_" + photo.getMetaData().getId() + ".jpg");
			updateDatabase();
			System.out.println("Upload complete.");
	
		} catch (IOException e) {
			System.out.println("Error while storing the picture on the local disk");
		}
		
	}

	public static void main(String args[]) {

		System.out.println("PhotoSharing RMI Server v0.1");
		System.out.println("(c) CONNECT Project - 2010");
		System.out.println("----");

		System.setProperty("java.security.policy", "resources/unsafe.policy");
		System.setProperty("java.rmi.server.codebase", "file://"+System.getProperty("user.dir")+"/bin/");
		//System.setProperty("java.rmi.server.codebase", "file://D:/Phd/Dev/eclipse/Workspace/photoSharing/bin/");
		
		System.setSecurityManager(new RMISecurityManager());
		try {
			RMIServer server = new RMIServer();
			LocateRegistry.createRegistry(2048);
			Registry r= LocateRegistry.getRegistry(2048);
			r.rebind(RMIServer.SERVER_ID, server);
			System.out.println("Server ready: Waiting for requests ...");

		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();

		} 
	}

}
